/*
 * Copyright (c) 2016-2018 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Mon Aug 22 2016
 */
#include "acmaneuver.h"

#include <sys/time.h>

#include <fcntl.h>
#include <unistd.h>

#include <cstdio>

#include "maneuver_c_types.h"
#include "codels.h"


/* --- Activity take_off ------------------------------------------------ */

/** Validation codel mv_plan_cancel of activity take_off.
 *
 * Returns genom_ok.
 * Throws maneuver_e_nostate, maneuver_e_limits.
 */
genom_event
mv_plan_cancel(maneuver_ids_trajectory_t *trajectory,
               maneuver_configuration_s *reference,
               const genom_context self)
{
  /* (re)start from current state if a trajectory is being executed */
  if (trajectory->t._length > 0) {
    *reference = trajectory->t._buffer[trajectory->i];
    trajectory->t._length = 0;
    trajectory->i = 0;
  }

  return genom_ok;
}


/* --- Activity goto ---------------------------------------------------- */

/** Validation codel mv_plan_cancel of activity goto.
 *
 * Returns genom_ok.
 * Throws maneuver_e_nostate, maneuver_e_limits.
 */
/* already defined in service take_off validation */



/* --- Function velocity ------------------------------------------------ */

/** Validation codel mv_plan_cancel of function velocity.
 *
 * Returns genom_ok.
 * Throws maneuver_e_limits.
 */
/* already defined in service take_off validation */



/* --- Activity replay -------------------------------------------------- */

/** Validation codel mv_plan_cancel of activity replay.
 *
 * Returns genom_ok.
 * Throws maneuver_e_nostate, maneuver_e_limits, maneuver_e_sys.
 */
/* already defined in service take_off validation */



/* --- Activity stop ---------------------------------------------------- */

/** Validation codel mv_plan_cancel of activity stop.
 *
 * Returns genom_ok.
 * Throws .
 */
/* already defined in service take_off validation */



/* --- Function set_bounds ---------------------------------------------- */

/** Codel mv_set_bounds of function set_bounds.
 *
 * Returns genom_ok.
 */
genom_event
mv_set_bounds(maneuver_planner_s **planner, double xmin, double xmax,
              double ymin, double ymax, double zmin, double zmax,
              double yawmin, double yawmax, const genom_context self)
{
  (*planner)->robot.getDof(0).setPositionMin(xmin);
  (*planner)->robot.getDof(0).setPositionMax(xmax);
  (*planner)->robot.getDof(1).setPositionMin(ymin);
  (*planner)->robot.getDof(1).setPositionMax(ymax);
  (*planner)->robot.getDof(2).setPositionMin(zmin);
  (*planner)->robot.getDof(2).setPositionMax(zmax);
  (*planner)->robot.getDof(3).setPositionMin(yawmin);
  (*planner)->robot.getDof(3).setPositionMax(yawmax);
  return genom_ok;
}


/* --- Function set_velocity_limit -------------------------------------- */

/** Codel mv_set_velocity_limit of function set_velocity_limit.
 *
 * Returns genom_ok.
 * Throws maneuver_e_limits.
 */
genom_event
mv_set_velocity_limit(maneuver_planner_s **planner,
                      maneuver_planner_s **vplanner, double v,
                      double w, const genom_context self)
{
  int i;

  if (v < 0. || w < 0.) return maneuver_e_limits(self);

  for(i = 0; i < 3; i++)
    (*planner)->robot.getDof(i).setVelocityMax(v);
  (*planner)->robot.getDof(3).setVelocityMax(w);

  for(i = 0; i < 3; i++) {
    (*vplanner)->robot.getDof(i).setPositionMin(-v);
    (*vplanner)->robot.getDof(i).setPositionMax(v);
  }
  (*vplanner)->robot.getDof(3).setPositionMin(-w);
  (*vplanner)->robot.getDof(3).setPositionMax(w);

  return genom_ok;
}


/* --- Function set_acceleration_limit ---------------------------------- */

/** Codel mv_set_acceleration_limit of function set_acceleration_limit.
 *
 * Returns genom_ok.
 * Throws maneuver_e_limits.
 */
genom_event
mv_set_acceleration_limit(maneuver_planner_s **planner,
                          maneuver_planner_s **vplanner, double a,
                          double dw, const genom_context self)
{
  int i;

  if (a < 0. || dw < 0.) return maneuver_e_limits(self);

  for(i = 0; i < 3; i++)
    (*planner)->robot.getDof(i).setAccelerationMax(a);
  (*planner)->robot.getDof(3).setAccelerationMax(dw);

  for(i = 0; i < 3; i++)
    (*vplanner)->robot.getDof(i).setVelocityMax(a);
  (*vplanner)->robot.getDof(3).setVelocityMax(dw);

  return genom_ok;
}


/* --- Function set_jerk_limit ------------------------------------------ */

/** Codel mv_set_jerk_limit of function set_jerk_limit.
 *
 * Returns genom_ok.
 * Throws maneuver_e_limits.
 */
genom_event
mv_set_jerk_limit(maneuver_planner_s **planner,
                  maneuver_planner_s **vplanner, double j, double ddw,
                  const genom_context self)
{
  int i;

  if (j < 0. || ddw < 0.) return maneuver_e_limits(self);

  for(i = 0; i < 3; i++)
    (*planner)->robot.getDof(i).setJerkMax(j);
  (*planner)->robot.getDof(3).setJerkMax(ddw);

  for(i = 0; i < 3; i++)
    (*vplanner)->robot.getDof(i).setAccelerationMax(j);
  (*vplanner)->robot.getDof(3).setAccelerationMax(ddw);

  return genom_ok;
}


/* --- Function set_snap_limit ------------------------------------------ */

/** Codel mv_set_snap_limit of function set_snap_limit.
 *
 * Returns genom_ok.
 * Throws maneuver_e_limits.
 */
genom_event
mv_set_snap_limit(maneuver_planner_s **planner,
                  maneuver_planner_s **vplanner, double s, double dddw,
                  const genom_context self)
{
  int i;

  if (s < 0. || dddw < 0.) return maneuver_e_limits(self);

  for(i = 0; i < 3; i++)
    (*planner)->robot.getDof(i).setSnapMax(s);
  (*planner)->robot.getDof(3).setSnapMax(dddw);

  for(i = 0; i < 3; i++) {
    (*vplanner)->robot.getDof(i).setJerkMax(s);
    (*vplanner)->robot.getDof(i).setSnapMax(1000. * s);
  }
  (*vplanner)->robot.getDof(3).setJerkMax(dddw);
  (*vplanner)->robot.getDof(3).setSnapMax(1000. * dddw);

  return genom_ok;
}


/* --- Function get_limits ---------------------------------------------- */

/** Codel mv_get_limits of function get_limits.
 *
 * Returns genom_ok.
 */
genom_event
mv_get_limits(maneuver_planner_s **planner, double *xmin, double *xmax,
              double *ymin, double *ymax, double *zmin, double *zmax,
              double *yawmin, double *yawmax, double *v, double *w,
              double *a, double *dw, double *j, double *ddw, double *s,
              double *dddw, const genom_context self)
{
  *xmin = (*planner)->robot.getDof(0).getPositionMin();
  *xmax = (*planner)->robot.getDof(0).getPositionMax();
  *ymin = (*planner)->robot.getDof(1).getPositionMin();
  *ymax = (*planner)->robot.getDof(1).getPositionMax();
  *zmin = (*planner)->robot.getDof(2).getPositionMin();
  *zmax = (*planner)->robot.getDof(2).getPositionMax();
  *yawmin = (*planner)->robot.getDof(3).getPositionMin();
  *yawmax = (*planner)->robot.getDof(3).getPositionMax();

  *v = (*planner)->robot.getDof(0).getVelocityMax();
  *w = (*planner)->robot.getDof(3).getVelocityMax();

  *a = (*planner)->robot.getDof(0).getAccelerationMax();
  *dw = (*planner)->robot.getDof(3).getAccelerationMax();

  *j = (*planner)->robot.getDof(0).getJerkMax();
  *ddw = (*planner)->robot.getDof(3).getJerkMax();

  *s = (*planner)->robot.getDof(0).getSnapMax();
  *dddw = (*planner)->robot.getDof(3).getSnapMax();

  return genom_ok;
}


/* --- Function velocity ------------------------------------------------ */

/** Codel mv_plan_velocity of function velocity.
 *
 * Returns genom_ok.
 * Throws maneuver_e_limits.
 */
genom_event
mv_plan_velocity(const maneuver_planner_s *vplanner,
                 const maneuver_configuration_s *reference, double vx,
                 double vy, double vz, double wz, double ax, double ay,
                 double az, double duration,
                 sequence_maneuver_configuration_s *path,
                 const genom_context self)
{
  struct timeval tv1, tv2;
  gettimeofday(&tv1, NULL);

  kdtp::State from(vplanner->robot);
  kdtp::State to(vplanner->robot);
  genom_event e;

  /* uses the planner for velocity: position() is actually velocity etc. */
  from.position()[0] = reference->vel[0];
  from.position()[1] = reference->vel[1];
  from.position()[2] = reference->vel[2];
  from.position()[3] = reference->vel[5];

  from.velocity()[0] = reference->acc[0];
  from.velocity()[1] = reference->acc[1];
  from.velocity()[2] = reference->acc[2];
  from.velocity()[3] = reference->acc[5];

  from.acceleration()[0] = reference->jer[0];
  from.acceleration()[1] = reference->jer[1];
  from.acceleration()[2] = reference->jer[2];
  from.acceleration()[3] = reference->jer[5];

  to.position()[0] = vx;
  to.position()[1] = vy;
  to.position()[2] = vz;
  to.position()[3] = wz;

  to.velocity()[0] = ax;
  to.velocity()[1] = ay;
  to.velocity()[2] = az;

  kdtp::LocalPath lpath(vplanner->robot, from, to, duration);

  e = mv_check_duration(lpath, duration, self);
  if (e) return e;
  e = mv_sample_velocity(reference->pos, lpath, path, self);
  if (e) return e;

  gettimeofday(&tv2, NULL);
  printf("%g\n", tv2.tv_sec - tv1.tv_sec + 1e-6*(tv2.tv_usec - tv1.tv_usec));

  return genom_ok;
}

/** Codel mv_push_path of function velocity.
 *
 * Returns genom_ok.
 * Throws maneuver_e_limits.
 */
genom_event
mv_push_path(const sequence_maneuver_configuration_s *path,
             maneuver_configuration_s *reference,
             maneuver_ids_trajectory_t *trajectory,
             const genom_context self)
{
  size_t i, l;

  if (!path->_length) return genom_ok;

  l = trajectory->t._length + path->_length;
  if (trajectory->t._maximum < l || trajectory->t._maximum > 2 * l)
    if (genom_sequence_reserve(&trajectory->t, l))
      return mv_e_sys_error(NULL, self);

  for(i = 0; i < path->_length; i++)
    trajectory->t._buffer[trajectory->t._length++] = path->_buffer[i];

  *reference = path->_buffer[path->_length-1];

  return genom_ok;
}


/* --- Function log ----------------------------------------------------- */

/** Codel mv_log of function log.
 *
 * Returns genom_ok.
 * Throws maneuver_e_sys.
 */
genom_event
mv_log(const char path[64], uint32_t decimation, maneuver_log_s **log,
       const genom_context self)
{
  int fd;

  fd = open(path, O_WRONLY|O_APPEND|O_CREAT|O_TRUNC, 0666);
  if (fd < 0) return mv_e_sys_error(path, self);

  if (write(fd, mv_log_header_fmt "\n", sizeof(mv_log_header_fmt)) < 0)
    return mv_e_sys_error(path, self);

  if ((*log)->req.aio_fildes >= 0) {
    close((*log)->req.aio_fildes);

    if ((*log)->pending)
      while (aio_error(&(*log)->req) == EINPROGRESS)
        /* empty body */;
  }
  (*log)->req.aio_fildes = fd;
  (*log)->pending = false;
  (*log)->skipped = false;
  (*log)->decimation = decimation < 1 ? 1 : decimation;
  (*log)->missed = 0;
  (*log)->total = 0;

  return genom_ok;
}


/* --- Function log_stop ------------------------------------------------ */

/** Codel mv_log_stop of function log_stop.
 *
 * Returns genom_ok.
 */
genom_event
mv_log_stop(maneuver_log_s **log, const genom_context self)
{
  if (*log && (*log)->req.aio_fildes >= 0)
    close((*log)->req.aio_fildes);
  (*log)->req.aio_fildes = -1;

  return genom_ok;
}


/* --- Function log_info ------------------------------------------------ */

/** Codel mv_log_info of function log_info.
 *
 * Returns genom_ok.
 */
genom_event
mv_log_info(const maneuver_log_s *log, uint32_t *miss, uint32_t *total,
            const genom_context self)
{
  *miss = *total = 0;
  if (log) {
    *miss = log->missed;
    *total = log->total;
  }

  return genom_ok;
}
